/****************************************************************************
*									    *	
*	mpi_sort.c                                                           *
*									    *
* ------------------------------------------------------------------------- *
*									    * 
*	PARALLELES SORTIEREN EINES ZAHLENVEKTORS !!!                        *
*									    *
*       Dieses Programm (Prozess 0) verteilt den Vektor gleichmaessig       *
*       auf die einzelnen Prozessoren. Nun wird lokal auf allen Pro-        *
*       zessoren parallel mit einem Quicksort sortiert. Das Zusammen-       *
*       fuegen der Teilergebnisse erfolgt anschliessend mittels             *
*       sogenannter Sequential Merge-Schritte.                              *
*       Parallelisierung erfolgt mit Hilfe des MPI`s.                       *
*
****************************************************************************/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <mpi.h>
#include "mymatrix.h"

typedef int bool;
#define true 1
#define false 0


/* Funktionsprototypen */
void merge_split(double *pVector, int n, int iMyRank, int iNbrProcs);  
void sequential_merge(double *tmpA, double *tmpB, int iEndA, int iEndB, 
                      bool bComputeLowerPart);
void quick_sort(double *vectorpart, int left, int right);

/* Hilfsvektor, wird eigentlich nur in sequential Merge benoetigt */
/* ist nur global definiert, um ein staendiges allokieren/deallokieren bei jedem 
 * Aufruf zu vermeiden */
double *pVecBuffer;    

int main(int argc, char *argv[]) {
  int n, test,             /* Groesse des Vektors bzw. Anzahl der Elemente */ 
    iMyRank,               /* Prozessornummer */
    iNbrProcs,             /* Anzahl der verfuegbaren Prozessoren */
    iError,                /* Fehlervariable */
    i,j;                   /* Zaehler */
  double **ppTemp;                  	/* zwischenprodukt zum erzeugen/laden
					/* des Vektors */
  double *pVector = NULL;               /* zu berechnender Vektor */
  double *time_start, *time_end;
  char *szInputFile, *szOutputFile;     /* Datei-Hilfsvariable */
  int hInputFile, hOutputFile;
  int runs = 10;                        /* Anzahl der Durchlaeufe */
  bool bGenerateRandom = 0;
  double time_average,
    time_best, 
    time_worst;     /* Durchschnittliche, beste und schlechteste Zeit bei der Zeitmessung*/
  
  /* Initialisierung der Matrix Bibilothek */
  matInit();
  
  /* Initialisierung des MPI`s */
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &iMyRank);
  MPI_Comm_size(MPI_COMM_WORLD, &iNbrProcs);

 /* Test auf Korrekte Kommandozeilenparameter */
  if (argc != 3) {
    if (iMyRank == 0) {
      fprintf(stderr, "benutzung: %s eingabedatei ausgabedatei\n",argv[0]);
      fprintf(stderr, "oder:      %s vektorgroesse testdurchlaeufe\n",argv[0]);
    }
    matFinalize();
    MPI_Abort(MPI_COMM_WORLD, -1);
  }
  if (atoi(argv[1]) <= 0) {
    szInputFile = argv[1];
    szOutputFile = argv[2];
    if (matReadDims(&test, &n, szInputFile) != MYMATRIX_ERR_OK) {
      if (iMyRank == 0)
	fprintf(stderr, "Die Matrix %s konnte nicht geoeffnet werden\n",
		szInputFile);
      matFinalize();
      MPI_Abort(MPI_COMM_WORLD, -1);
    }
    if ((test != 1) || (n < 1)) {
      if (iMyRank == 0)
	fprintf(stderr, "Die Matrix hat nicht das Format 1 x n!\n");
      matFinalize();
      MPI_Abort(MPI_COMM_WORLD, -1);
    }
    bGenerateRandom = false;
    runs = 1;
  }
  else {
    n = atoi(argv[1]);
    if (n < 1) {
      if (iMyRank == 0)
	fprintf(stderr, "Der Vektor muss mindestens die Groesse 1 haben!\n");
      matFinalize();
      MPI_Abort(MPI_COMM_WORLD, -1);
      exit(-1); 
    }
    runs = atoi(argv[2]);
    if (runs < 1)
      runs = 10;
    bGenerateRandom = true;
  }
  
  /* Einlesen oder generieren erfolgt auf Prozessor 0 (root) */  
  if ( iMyRank == 0 ) {
    if(bGenerateRandom) {
      /* erzeugen */
      ppTemp = matMalloc(1, n);
      matGenerateRandom(ppTemp);  
      printf( "Vektor wurde durch Zufallsdaten erzeugt...\n" ); fflush(stdout);
    }
    else {
      /* Dateien oeffnen */
      hInputFile = matRead( &ppTemp, szInputFile);
      if ( hInputFile != 0 ) {    
	fprintf(stderr, "Eingabedatei %s konnte nicht geoeffnet werden\n", szInputFile);
	printf("Programmabruch.\n");
	matFinalize();
	MPI_Abort(MPI_COMM_WORLD, -1);
      }	    
      printf( "Vektor wurde  aus Datei gelesen...\n" ); fflush(stdout);
    }
    
    /* Zeiger auf den Datenteil der Matrix als Zeiger auf den Vektor benutzen */
    pVector = ppTemp[0];
  } 
  
  time_start = (double*)malloc( runs * sizeof(double) );
  time_end = (double*)malloc( runs * sizeof(double) );
  pVecBuffer = (double *) malloc( (n/iNbrProcs +1) * sizeof( double ));
  
  for( i=0; i < runs; i++ ) {
    if(iMyRank == 0) {
      time_start[i] = MPI_Wtime();
    }
    
    /* Aufruf des Merge-Split-Algorithmus */         
    merge_split(pVector, n, iMyRank, iNbrProcs);
    
    if(iMyRank == 0) {
      time_end[i] = MPI_Wtime();
      
      for( j = 1; j < n; j++ )
	if( pVector[j-1] > pVector[j] )
	  printf("Vector wurde an der %d. Position falsch sortiert (%f ist nicht >= %f)!!! \n",
		 j, pVector[j], pVector[j-1] );
      if(bGenerateRandom) {
	matGenerateRandom(ppTemp);
      }
    }
    MPI_Barrier(MPI_COMM_WORLD);
  }
  
  if ( iMyRank == 0 ) { 
    if( bGenerateRandom ) {
      time_average = 0;
      /* berechnung des durchschnittlichen Zeitbedarfs */
      for(i = 0; i < runs; i++)
	time_average += (time_end[i] - time_start[i]);
      time_average = time_average/runs;
      
      printf("\nParalleles Sortieren von %d Elementen\n", n);
      time_best = time_end[0] - time_start[0];
      time_worst = time_end[0] - time_start[0];
      for ( i=0; i < runs; i++ ) {
	if ( (time_end[i] - time_start[i]) < time_best )
	  time_best = time_end[i] - time_start[i];
	if ( (time_end[i] - time_start[i]) > time_worst )
	  time_worst = time_end[i] - time_start[i];
      }
      printf(" Beste Zeit             = %.6f s\n", time_best);
      printf(" Durchschnittliche Zeit = %.6f s\n", time_average);
      printf(" Schlechteste Zeit      = %.6f s\n", time_worst);
      
    } 
    else {
      time_average = time_end[0] - time_start[0];
      printf("\nParalleles Sortieren von %d Elementen\n", n);
      printf(" Benoetigte Zeit = %.6f s\n", time_average);
      if( matWrite(ppTemp, argv[2]) != 0 )
	printf("Schreiben der Datei fehlgeschlagen\n");
      matFree(ppTemp);
    }
  }
  
  /* Speicher freigeben */
  free(pVecBuffer);
  free(time_start);
  free(time_end);
  
  matFree(ppTemp);
  
  /* Smi Bereich verlassen */
  MPI_Finalize();

  /* Matrix Modul abschliessen */
  matFinalize();  
  
  return(0);
}



/*
 * paralleler Teil: merge-split
 */
void merge_split(double *pVector, int n, int iMyRank, int iNbrProcs) {
  int *pStart,                /* Verteilung der Zeilen auf die Prozessoren */
    rest,                     /* restliche Zeilen */
    iSizeB,                   /* Laenge des Feldes tmpB */
    i;  		      /* Zaehler */
  double *tmpA, *tmpB;        /* Hilfsvektoren */
  
  MPI_Status status;          /* fuer MPI_Recv */
  
  /* Berechnung der Vektorverteilung */
  pStart = (int *) malloc( (iNbrProcs+1) * sizeof(int) );
  pStart[0] = 0;
  rest = n % iNbrProcs;
  for ( i = 1; i <= iNbrProcs; i++ ) {
    if ( rest != 0 ) {
      pStart[i] = pStart[i-1] + n/iNbrProcs + 1;
      rest--;
    } 
    else {
      pStart[i] = pStart[i-1] + n/iNbrProcs;
    }
  }
  
  /* Verschicken der jeweiligen Teilvektoren */
  if ( iMyRank == 0 ) {
    for ( i = 1; i<iNbrProcs; i++ ) {
      MPI_Send(&pVector[pStart[i]], pStart[i+1]-pStart[i], MPI_DOUBLE, i, 2, MPI_COMM_WORLD);
    }
    tmpA = (double *) malloc( (pStart[iMyRank+1] - pStart[iMyRank]) * sizeof(double) );
    for ( i = 0; i<pStart[1]-pStart[0]; i++ ) 
      tmpA[i]=pVector[i];
  } 
  else {
    tmpA = (double *) malloc( (pStart[iMyRank+1] - pStart[iMyRank]) * sizeof(double) );
    MPI_Recv(tmpA, pStart[iMyRank+1] - pStart[iMyRank], MPI_DOUBLE, 0, 2, MPI_COMM_WORLD, &status); 
  }
  tmpB = (double *) malloc( (pStart[1] - pStart[0]) * sizeof(double) );
  
  /* paralleles Sortieren auf allen Prozessoren */ 
  quick_sort(tmpA, 0, pStart[iMyRank+1]-pStart[iMyRank]-1);
  
  /* Anzahl der Schleifendurchlaeufe ist abhaengig von
     der Prozessoranzahl; ob gerade oder ungerade */
  
  /* Teilfolgen mit Sequential-Merge sortieren */
  for ( i = 0 ; i < iNbrProcs/2 + (iNbrProcs & 1); i++ ) {
    /* Austausch der Daten */
    /* von geraden nach ungeraden Prozessoren (und zurueck) */
    if ( !( iMyRank % 2 ) ) { 
      if ( iMyRank+1 < iNbrProcs ) {
	/* bei ungerader Prozessoranzahl ist der */
	/* letzte Prozessor natuerlich gerade und darf nicht senden */
	MPI_Send(tmpA, pStart[iMyRank+1]-pStart[iMyRank], MPI_DOUBLE, iMyRank+1, 2, MPI_COMM_WORLD);          
	
	iSizeB = pStart[iMyRank+2]-pStart[iMyRank+1];
	MPI_Recv(tmpB, iSizeB, MPI_DOUBLE, iMyRank+1, 2, MPI_COMM_WORLD, &status);
	
	sequential_merge(tmpA, tmpB, pStart[iMyRank+1]-pStart[iMyRank]-1, iSizeB-1, true);
      }
    } 
    else {
      iSizeB = pStart[iMyRank]-pStart[iMyRank-1]; 
      MPI_Recv(tmpB, iSizeB, MPI_DOUBLE, iMyRank-1, 2, MPI_COMM_WORLD, &status);   
      
      MPI_Send(tmpA, pStart[iMyRank+1]-pStart[iMyRank], MPI_DOUBLE, iMyRank-1, 2, MPI_COMM_WORLD);
	
      sequential_merge(tmpA, tmpB, pStart[iMyRank+1]-pStart[iMyRank]-1, iSizeB-1, false);      
    }
    
    /* von ungeraden nach geraden Prozessoren (und zurueck) */
    if ( iMyRank % 2 ) { 
      if( iMyRank + 1 < iNbrProcs ) {
	/* bei gerader Prozessoranzahl ist der */
	/* letzte Prozessor natuerlich ungerade und darf nicht senden */
	MPI_Send(tmpA, pStart[iMyRank+1]-pStart[iMyRank], MPI_DOUBLE, iMyRank+1, 2, MPI_COMM_WORLD);          
	
	iSizeB = pStart[iMyRank+2]-pStart[iMyRank+1];
	MPI_Recv(tmpB, iSizeB, MPI_DOUBLE, iMyRank+1, 2, MPI_COMM_WORLD, &status);
	
	sequential_merge(tmpA, tmpB, pStart[iMyRank+1]-pStart[iMyRank]-1, iSizeB-1, true);
      }
    } 
    else {
      if(iMyRank != 0) {    
	/* Prozess 0 darf nicht empfangen */
	iSizeB = pStart[iMyRank]-pStart[iMyRank-1]; 
	MPI_Recv(tmpB, iSizeB, MPI_DOUBLE, iMyRank-1, 2, MPI_COMM_WORLD, &status);   
	
	MPI_Send(tmpA, pStart[iMyRank+1]-pStart[iMyRank], MPI_DOUBLE, iMyRank-1, 2, MPI_COMM_WORLD);
	
	sequential_merge(tmpA, tmpB, pStart[iMyRank+1]-pStart[iMyRank]-1, iSizeB-1, false);
	
      }
    }      
  } 
  
  /* Teilvektoren wieder an Prozessor 0 schicken - MPI_Gather ist nicht nutzbar, da die 
     Teilvektoren nicht alle die gleiche Laenge haben */
  if (iMyRank != 0) {
    MPI_Send(tmpA, pStart[iMyRank+1]-pStart[iMyRank], MPI_DOUBLE, 0, 2, MPI_COMM_WORLD);
  } 
    else {
      for (i = 1; i < iNbrProcs; i++) {
	MPI_Recv(&pVector[pStart[i]], pStart[i+1]-pStart[i], MPI_DOUBLE, i, 2, MPI_COMM_WORLD, &status);
      }
      for ( i=0; i<pStart[iMyRank+1]; i++ )
	pVector[i] = tmpA[i];
    }
  
  free(tmpA);
  free(tmpB);        
    
  return;
}



/*
 * alternativ kann auch der quicksort aus der clib verwendet werden!
 */
void quick_sort(double *vectorpart, int left, int right) {
  int i = left, j = right;                    /* linke und rechte Grenze */
  double x, y;                                  /* Pivot- und Hilfselement */
  
  /* Pivotelement bestimmen */
  x = vectorpart[(left+right)/2];
  
  /* Vektor von links nach rechts und umgekehrt durchlaufen
     und eventuell gefundene nichtpassende Elemente vertauschen */
  do {
    while ( vectorpart[i] < x )
      i++;
    
    while ( x < vectorpart[j] )
      j--;
    
    if ( i <= j ) {
      y = vectorpart[i];
      vectorpart[i] = vectorpart[j];
      vectorpart[j] = y;
      i++;
      j--;
    }             
  } while ( i <= j );
  
  /* Aufruf von Quick-Sort fuer die linke bzw. rechte Seite */
  if ( left < j )
    quick_sort( vectorpart, left, j );
  
  if ( i < right )
    quick_sort( vectorpart, i, right );
  
  return;     
}



/*
 * lokales Mischen der sortierten Teilvektoren
 */

void sequential_merge(double *tmpA, double *tmpB, int iEndA, int iEndB, 
                      bool bComputeLowerPart) 
{
  /* Zaehler */
  int i = 0, 
    j = 0,
    k;
  int groesse;
  
  groesse  = iEndA + 1;                                         
  
  /* die untere Haelfte wird bearbeitet */ 
  if( bComputeLowerPart ) {
    for (k = 0; k < (iEndA+1); k++) 
    {
      if ( i <= iEndA && j <= iEndB ) {
	/* die kleinsten Werte werden aufsteigend sortiert ins 
	   Hilfsfeld geschrieben */ 
	if ( tmpA[i] < tmpB[j] ) {
	  pVecBuffer[k] = tmpA[i];
	  i++;
	} 
	else {
	  pVecBuffer[k] = tmpB[j];
	  j++;
	}
      } 
      else {
	/* falls ein Feld kleiner als das andere ist */
	if ( i > iEndA ) {
	  pVecBuffer[k] = tmpB[j];
	  j++;
	}
	else 
	{
	  pVecBuffer[k] = tmpA[i];
	  i++;
	}    
      }
    } 
  } 
  else {
    /* die obere Haelfte wird bearbeitet */ 
    i = iEndA;
    j = iEndB;
    
    for( k = iEndA; k >= 0; k-- ) {
      /* die groessten Werte werden abfallend sortiert ins 
	 Hilfsfeld geschrieben */         
      if ( i >= 0 && j >= 0 ) {
	if ( tmpA[i] > tmpB[j] ) {
	  pVecBuffer[k] = tmpA[i];
	  i--;
	}
	else {
	  pVecBuffer[k] = tmpB[j];
	  j--;
	}
      } 
      else {
	/* falls ein Feld kleiner als das andere ist */
	if ( i < 0 ) {
	  pVecBuffer[k] = tmpB[j];
	  j--;
	} 
	else {
	  pVecBuffer[k] = tmpA[i];
	  i--;
	}    
      }
    }
  }  
  /* die sortierten Daten werden wieder im Feld A abgespeichert */    
  for ( i = 0; i <= iEndA; i++)
    tmpA[i] = pVecBuffer[i];  
  
  return;
}          







